"
 (C) 2011 by Holger Hans Peter Freyther
 All Rights Reserved

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"

PackageLoader
    fileInPackage: 'OsmoCore';
    fileInPackage: 'Sockets'.

Object subclass: SIPTransaction [
    | sem useragent initial_dialog state timeout success failure notification
      cseq branch retransmit_time fail_time removal |

    SIPTransaction class >> stateInitial    [ <category: 'states'> ^ #initial ]
    SIPTransaction class >> stateTrying     [ <category: 'states'> ^ #trying ]
    SIPTransaction class >> stateProceeding [ <category: 'states'> ^ #proceeding ]
    SIPTransaction class >> stateCompleted  [ <category: 'states'> ^ #completed ]
    SIPTransaction class >> stateTerminated [ <category: 'states'> ^ #terminated ]
    SIPTransaction class >> stateCanceled   [ <category: 'states'> ^ #canceled ]

    SIPTransaction class >> createWith: aDialog on: aUA  cseq: aCseq [
        <category: 'creation'>
        ^ self new
            initialize;
            userAgent: aUA;
            initialDialog: aDialog;
            setupTransaction: aCseq;
            yourself.
    ]

    initialize [
        <category: 'creation'>
        sem := RecursionLock new.
    ]

    initialDialog: aDialog [
        <category: 'creation'>
        initial_dialog := aDialog
    ]

    userAgent: aUA [
        <category: 'creation'>
        useragent := aUA
    ]

    state [
        <category: 'state'>
        ^ state ifNil: [^ self class stateInitial]
    ]

    changeState: aState [
        <category: 'state'>

        "I could verify the state change here"
        sem critical: [
            state := aState.
        ].
    ]

    onTimeout: aTimeout [
        <category: 'setup'>
        timeout := aTimeout
    ]

    onSuccess: aSuc [
        <category: 'setup'>
        success := aSuc
    ]

    onFailure: aFail [
        <category: 'setup'>
        failure := aFail
    ]

    onNotification: aNot [
        <category: 'setup'>
        notification := aNot
    ]

    setupTransaction: aCseq [
        <category: 'private'>
        "I setup the transaction"
        useragent addTransaction: self.
        cseq := aCseq.
        branch := useragent class generateBranch.
    ]

    queueData: aData dialog: aDialog [
        | res datagram |
        <category: 'private'>

        "validate the output"
        res := SIPGrammar parse: aData.

        datagram := Sockets.Datagram
                                data: aData
                                address: (Sockets.SocketAddress
                                            byName: aDialog destIp)
                                port: aDialog destPort.
        useragent queueData: datagram.
    ]

    branch [
        <category: 'accessing'>
        ^ branch
    ]

    respNotification: aReq dialog: aDialog [
        <category: 'private-dispatch'>

        self changeState: self class stateProceeding.

        notification ifNotNil: [
            notification value: aReq value: aDialog]
    ]

    respSuccess: aReq dialog: aDialog [
        <category: 'private-dispatch'>

        self removeTransaction.
        ^ success isNil
                ifTrue: [true]
                ifFalse: [success value: aReq value: aDialog].
    ]

    respFailure: aReq dialog: aDialog [
        <category: 'private-dispatch'>

        self removeTransaction.
        failure ifNotNil: [
            failure value: aReq value: aDialog]
    ]

    checkSequenceNumber: aReq [
        | new_cseq |
        <category: 'private-dispatch'>
        "I have to verify the sequence numbers..."
        new_cseq := (aReq parameter: 'CSeq' ifAbsent: [
            self logError: '%1(%2) response lacks CSeq.'
                    % {self class. self branch} area: #sip.
            ^ false
        ]) number.

        cseq = new_cseq ifFalse: [
            self logError: '%1(%2) wrong cseq:. %1 %2.'
                    % {self class. self branch. cseq. new_cseq} area: #sip.
            ^ false
        ].

        ^ true
    ]


    stopRetransmitTimer [
        <category: 'private'>
        "I stop the retransmit timers, e.g. because there was a request,
         or there was a timeout, or someone canceled things."
        sem critical: [
            retransmit_time cancel.
            fail_time cancel.
        ]
    ]

    newData: aResp [
        | dialog |
        <category: 'private-dispatch'>

        ('Found response %1 %2' % {aResp code. aResp phrase}) printNl.

        (self checkSequenceNumber: aResp) ifFalse: [^ false].

        "Store all dialogs.... and match them..."
        dialog := initial_dialog newFromRequest: aResp.

        self stopRetransmitTimer.
        self dispatchDialog: dialog response: aResp.
    ]

    dispatchDialog: dialog response: aResp [
        | code |
        <category: 'private-dispatch'>

        code := aResp code asInteger.
        code < 200 ifTrue: [
            ^ self respNotification: aResp dialog: dialog
        ].

        code = 200 ifTrue: [
            ^ self respSuccess: aResp dialog: dialog
        ].

        code > 200 ifTrue: [
            ^ self respFailure: aResp dialog: dialog
        ]
    ]

    start [
        sem critical: [
            self state = self class stateInitial ifFalse: [
                ^ self error: 'Can not restart.'
            ].

            "Enter the state, remember the timeout"
            self changeState: self class stateTrying.

            fail_time := Osmo.TimerScheduler instance
                            scheduleInSeconds: 60 block:
                                [sem critical: [self timedout]].
            self retransmit.
        ]
    ]

    retransmit [
        <category: 'timeout'>
        retransmit_time := Osmo.TimerScheduler instance
                            scheduleInSeconds: 10 block:
                                [sem critical: [self retransmit]].
        self transmit.
    ]

    timedout [
        <category: 'timeout'>
        useragent removeTransaction: self.
        retransmit_time ifNotNil: [retransmit_time cancel].
        timeout value.
    ]

    removeTransaction [
        <category: 'timeout'>
        "We want to remove things but we might get a late reply due a packet loss."
        removal ifNotNil: [removal cancel].
        removal := Osmo.TimerScheduler instance
                        scheduleInSeconds: 10 block:
                            [sem critical: [useragent removeTransaction: self]].
    ]

    createInvite: sdp [
        | invite |
        <category: 'invite'>

        invite := (SIPInviteRequest from: initial_dialog)
                        sdp: sdp;
                        addParameter: 'Via' value: (useragent generateVia: branch);
                        addParameter: 'CSeq' value: '%1 %2' % {cseq. 'INVITE'};
                        addParameter: 'Allow' value: 'ACK,BYE';
                        addParameter: 'Call-ID' value: initial_dialog callId;
                        yourself.
        useragent injectDefaults: invite.
        ^ invite
    ]

    createAck: aBranch dialog: aDialog [
        | ack |
        <category: 'ack'>

        ack := (SIPACKRequest from: aDialog)
            addParameter: 'Via' value: (useragent generateVia: aBranch);
            addParameter: 'CSeq' value: '%1 %2' % {cseq. 'ACK'};
            addParameter: 'Call-ID' value: aDialog callId;
            yourself.
        useragent injectDefaults: ack.
        ^ ack
    ]

    createBye: aDialog [
        | bye |
        <category: 'invite'>

        bye := (SIPByeRequest from: aDialog)
                        addParameter: 'Via' value: (useragent generateVia: branch);
                        addParameter: 'CSeq' value: '%1 %2' % {cseq. 'BYE'};
                        addParameter: 'Call-ID' value: aDialog callId;
                        yourself.
        useragent injectDefaults: bye.
        ^ bye
    ]

    createCancel [
        | cancel |
        <category: 'cancel'>

        cancel := (SIPCancelRequest from: initial_dialog)
                        addParameter: 'Via' value: (useragent generateVia: branch);
                        addParameter: 'CSeq' value: '%1 %2' % {cseq. 'CANCEL'};
                        addParameter: 'Call-ID' value: initial_dialog callId;
                        yourself.
        useragent injectDefaults: cancel.
        ^ cancel.
    ]
]

SIPTransaction subclass: SIPInviteTransaction [
    | sdp ack_branch canceled |
    <category: 'RFC3161 17.2.1'>

    "200ms to get TRYING or OK"

    SIPInviteTransaction class >> createWith: aDialog on: aUA with: aSDP cseq: aCseq[
        <category: 'creation'>
        ^ (super createWith: aDialog on: aUA cseq: aCseq)
            instVarNamed: #sdp put: aSDP;
            yourself.
    ]

    initialize [
        <category: 'creation'>
        super initialize.
        canceled := false
    ]

    transmit [
        | invite |
        <category: 'transmit'>
        invite := self createInvite: sdp.
        self queueData: invite asDatagram dialog: initial_dialog.
    ]

    dispatchDialog: aDialog response: aResponse [
        | cseq |
        <category: 'private-dispatch'>
        "The INVITE transaction is a bit more complicated. It is the only
         transaction that can be canceled and we will need to do some things
         to check if this is canceled. We have some indirections here
         1.) we get a 200 for a BYE/CANCEL in the CSeq
         2.) we get a reply but we should cancel, e.g. we were waiting for
             a proceeding/notification or the call is too far in the setup
             and we will just bye it.
         3.) the normal dispatch..."

        cseq := aResponse parameter: 'CSeq' ifAbsent: [].
        cseq method = 'INVITE'
            ifTrue:  [self dispatchInvite: aDialog response: aResponse]
            ifFalse: [self dispatchOther: aDialog response: aResponse].
    ]

    dispatchOther: aDialog response: aResponse [
        | cseq code |
        <category: 'private-dispatch'>

        code := aResponse code asInteger.
        cseq := aResponse parameter: 'CSeq' ifAbsent: [].

        code = 200 ifTrue: [
            self removeTransaction.
            self queueData: (self createAck: branch dialog: aDialog) asDatagram
                    dialog: aDialog.
        ].
    ]

    handleCancel: aDialog response: aResponse [
        | code |
        <category: 'private-dispatch'>
        code := aResponse code asInteger.

        "We will send a CANCEL, maybe it is already the second."
        code < 200 ifTrue: [
            self queueData: (self createCancel asDatagram) dialog: initial_dialog.
        ].

        "We are connected but we didn't want to, let us BYE it"
        (code = 200 or: [(code > 200 and: [code ~= 487])]) ifTrue: [
            | bye branch |
            branch := useragent class generateBranch.
            self removeTransaction.
            self queueData: (self createAck: branch dialog: aDialog) asDatagram
                    dialog: aDialog.
            bye := SIPByeTransaction
                    createWith: aDialog on: useragent cseq: cseq + 1.
            bye start.
        ].
    ]

    dispatchInvite: aDialog response: aResponse [
        <category: 'private-dispatch'>
        "Send a cancel if this is a non final response"
        canceled
            ifTrue:  [self handleCancel: aDialog response: aResponse]
            ifFalse: [super dispatchDialog: aDialog response: aResponse].
    ]

    respSuccess: aReq dialog: aDialog [
        | branch |
        <category: 'private-dispatch'>

        branch := useragent class generateBranch.
        self changeState: self class stateTerminated.

        "We send the ACK, if our callbacks don't like the result we
         will need to send a BYE to stop that session."
        self queueData: (self createAck: branch dialog: aDialog) asDatagram dialog: aDialog.
        (super respSuccess: aReq dialog: aDialog) ifFalse: [| bye |
                bye := SIPByeTransaction
                        createWith: aDialog on: useragent cseq: cseq + 1.
                bye start.
        ]
    ]

    respFailure: aReq dialog: aDialog [
        <category: 'private-dispatch'>
        self changeState: self class stateCompleted.
        self queueData: (self createAck: branch dialog: aDialog) asDatagram
                dialog: aDialog.
        ^ super respFailure: aReq dialog: aDialog.
    ]

    cancel [
        <category: 'cancel'>
        canceled := true.
        self stopRetransmitTimer.
        self changeState: self class stateCanceled.

        self state = self class stateProceeding ifTrue: [
            self queueData: self createCancel asDatagram dialog: initial_dialog.
        ].
        self state = self class stateCompleted ifTrue: [
            self logError: 'SIPTransaction already completed.' area: #sip.
        ].
    ]
]

SIPTransaction subclass: SIPByeTransaction [
    <category: 'a bye...'>

    transmit [
        | bye |
        bye := self createBye: initial_dialog.
        self queueData: bye asDatagram dialog: initial_dialog.
    ]
]
